import asyncio as aio
import os
import time
import traceback
from asyncio import iscoroutinefunction
from copy import deepcopy
from functools import partial
from pathlib import Path
from types import MethodType

from ..context.session import SESSION_LOCAL, BotSessionManager
from ..meta import MODULE_MODE_FLAG, MODULE_MODE_SET
from ..types.abc import BotHookRunnerArgs, BotLife
from ..types.exceptions import *
from ..types.typing import *

if TYPE_CHECKING:
    from ..init import MeloBot
    from ..utils.config import BotConfig
    from ..utils.logger import Logger
    from .plugin import Plugin


class HookRunner:
    """
    bot 生命周期 hook 方法
    """

    def __init__(
        self,
        type: str,
        func: Callable[..., Coroutine[Any, Any, None]],
        plugin: Optional["Plugin"],
    ) -> None:
        self._func = func
        self._plugin = plugin
        # 对应：绑定的 func 是插件类的实例方法
        if plugin:
            self.cb = partial(self._func, plugin)
        # 对应：绑定的 func 是普通函数
        else:
            self.cb = self._func
        self.type = type


class BotHookBus:
    """
    bot 生命周期 hook 总线
    """

    __store__: Dict[BotLife, List[HookRunner]] = {
        v: [] for k, v in BotLife.__members__.items()
    }
    __logger: "Logger"

    @classmethod
    def _bind(cls, logger: "Logger") -> None:
        """
        初始化该类，绑定全局日志器和行为响应器
        """
        cls.__logger = logger

    @classmethod
    def _register(cls, hook_type: BotLife, runner: HookRunner) -> None:
        """
        注册一个生命周期运行器。由 plugin build 过程调用
        """
        if hook_type not in cls.__store__.keys():
            raise BotHookError(
                f"尝试添加一个生命周期 hook 方法，但是其指定的类型 {hook_type} 不存在"
            )
        cls.__store__[hook_type].append(runner)

    @classmethod
    def on(
        cls, hook_type: BotLife, callback: Callable[..., Coroutine[Any, Any, None]]
    ) -> None:
        """
        动态注册 hook 方法
        """
        if isinstance(callback, HookRunner):
            raise BotHookError("已注册的生命周期 hook 方法不能再注册")
        if not iscoroutinefunction(callback):
            raise BotHookError(f"生命周期 hook 方法 {callback.__name__} 必须为异步函数")
        if (
            isinstance(callback, partial) and isinstance(callback.func, MethodType)
        ) or isinstance(callback, MethodType):
            raise BotHookError("callback 应该是 function，而不是 bound method")
        runner = HookRunner(hook_type, callback, plugin=None)
        cls._register(hook_type, runner)

    @classmethod
    async def _run_on_ctx(cls, runner: HookRunner, *args, **kwargs) -> None:
        session = BotSessionManager.make_empty()
        token = SESSION_LOCAL._add_ctx(session)
        try:
            await runner.cb(*args, **kwargs)
        except Exception as e:
            e_name = e.__class__.__name__
            func_name = runner._func.__qualname__
            pre_str = "插件" + runner._plugin.ID if runner._plugin else "动态注册的"
            cls.__logger.error(
                f"{pre_str} hook 方法 {func_name} 发生异常：[{e_name}] {e}"
            )
            cls.__logger.debug(f"生命周期 hook 方法的 args: {args} kwargs：{kwargs}")
            cls.__logger.debug("异常回溯栈：\n" + traceback.format_exc().strip("\n"))
        finally:
            SESSION_LOCAL._del_ctx(token)

    @classmethod
    async def emit(
        cls, hook_type: BotLife, *args, wait: bool = False, **kwargs
    ) -> None:
        """
        触发一个生命周期事件。如果指定 wait 为 True，则会等待所有生命周期 hook 方法完成
        """
        if not wait:
            for runner in cls.__store__[hook_type]:
                aio.create_task(cls._run_on_ctx(runner, *args, **kwargs))
        else:
            for runner in cls.__store__[hook_type]:
                await cls._run_on_ctx(runner, *args, **kwargs)


class PluginProxy:
    """
    bot 插件代理类。供外部使用
    """

    def __init__(
        self,
        id: str,
        ver: str,
        path: Path,
        share_objs: list[tuple[str, str]],
        signal_methods: list[tuple[str, str]],
    ) -> None:
        self.id = id
        self.version = ver
        self.path = path
        self.shares = share_objs
        self.signal_methods = signal_methods


class BotProxy:
    """
    外部使用的 Bot 对象
    """

    def __init__(self) -> None:
        self.__origin__: "MeloBot"

    def _bind(self, bot_origin: "MeloBot") -> None:
        self.__origin__ = bot_origin

    @property
    def is_activate(self) -> bool:
        return not self.__origin__.slack

    def get_loop(self) -> aio.AbstractEventLoop:
        """
        获得运行 bot 的异步事件循环
        """
        return self.__origin__.loop

    def get_logger(self) -> "Logger":
        """
        获得 bot 全局日志器
        """
        return self.__origin__.logger

    def get_config(self) -> "BotConfig":
        """
        获得 bot 全局配置
        """
        return deepcopy(self.__origin__.config)

    def get_plugins(self) -> Dict[str, PluginProxy]:
        """
        获取 bot 当前所有插件信息
        """
        return {id: plugin._proxy for id, plugin in self.__origin__.plugins.items()}

    def slack(self) -> None:
        """
        使 bot 不再发送 action。但 ACTION_PRESEND 钩子依然会触发
        """
        self.__origin__.slack = True

    def activate(self) -> None:
        """
        使 bot 可以发送 action
        """
        self.__origin__.slack = False

    async def close(self) -> None:
        """
        关闭 bot
        """
        await self.__origin__.close()

    def is_module_run(self) -> bool:
        """
        检查是否是模块运行模式
        """
        return os.environ.get(MODULE_MODE_FLAG) == MODULE_MODE_SET

    async def restart(self) -> None:
        """
        重启 bot。只可在模块运行模式下使用
        """
        await self.__origin__.restart()

    def call_later(self, callback: partial, delay: float):
        """
        在指定的 delay 后调度一个 callback 执行。注意这个 callback 应该是同步方法
        """
        return self.get_loop.call_later(delay, callback)

    def call_at(self, callback: partial, timestamp: float):
        """
        在指定的时间戳后调度一个 callback 执行。注意这个 callback 应该是同步方法

        当 timestamp <= 当前时刻将立即执行
        """
        if timestamp <= time.time():
            return self.get_loop.call_soon(callback)
        else:
            return self.get_loop.call_later(timestamp - time.time(), callback)

    def async_later(
        self,
        callback: Coroutine[Any, Any, Any],
        delay: float,
        exit_cb: Coroutine[Any, Any, None] = None,
    ) -> aio.Future:
        """
        在指定的 delay 后调度一个 callback 执行。注意这个 callback 应该是协程。
        返回一个 future 对象，可用于等待结果。可以指定 exit_cb 用于在被迫 cancel 时调用
        """

        async def async_cb(fut: aio.Future) -> Any:
            try:
                await aio.sleep(delay)
                res = await callback
                fut.set_result(res)
            except aio.CancelledError:
                if exit_cb is not None:
                    await exit_cb
                callback.close()

        fut = aio.Future()
        aio.create_task(async_cb(fut))
        return fut

    def async_at(
        self,
        callback: Coroutine[Any, Any, Any],
        timestamp: float,
        exit_cb: Coroutine[Any, Any, None] = None,
    ) -> aio.Future:
        """
        在指定的 timestamp 调度一个 callback 执行。注意这个 callback 应该是协程。
        返回一个 future 对象，可用于等待结果。可以指定 exit_cb 用于在被迫 cancel 时调用

        当 timestamp <= 当前时刻将立即执行
        """
        if timestamp <= time.time():
            return self.async_later(callback, 0, exit_cb)
        else:
            return self.async_later(callback, timestamp - time.time(), exit_cb)

    def async_interval(
        self,
        cb_maker: Callable[[None], Coroutine[Any, Any, None]],
        interval: float,
        exit_cb: Coroutine[Any, Any, None] = None,
    ) -> aio.Task:
        """
        以指定的 interval 调度一个 callback 执行。cb_maker 是用于产生 callback 的协程产生器。
        返回一个 task 对象用于取消 callback。可以指定 exit_cb 用于在被迫 cancel 时调用
        """

        async def interval_cb():
            try:
                while True:
                    coro = cb_maker()
                    await aio.sleep(interval)
                    await coro
            except aio.CancelledError:
                if exit_cb is not None:
                    await exit_cb
                coro.close()

        t = aio.create_task(interval_cb())
        return t

    def on(
        self,
        hook_type: BotLife,
        callback: Callable[..., Coroutine[Any, Any, None]] = None,
    ):
        """
        用作装饰器，不传入 callback 参数，即可注册一个 bot 生命周期 hook 方法。

        也可直接调用此方法，传入 callback 参数来注册一个 hook 方法。
        callback 可以是类实例方法，也可以不是。callback 如果是类实例方法，请自行包裹为一个 partial 函数。

        例如你的插件类是：`A`，而你需要传递一个类实例方法：`A.xxx`，
        那么你的 callback 参数就应该是：`functools.partial(A.xxx, self)`
        """

        def make_args(
            func: Callable[..., Coroutine[Any, Any, None]]
        ) -> BotHookRunnerArgs:
            return BotHookRunnerArgs(func=func, type=hook_type)

        if callback is None:
            return make_args
        else:
            BotHookBus.on(hook_type, callback)

    def on_all_loaded(self):
        """
        用作装饰器，可注册一个 LOADED 生命周期 hook 方法
        """
        return self.on(BotLife.LOADED, None)

    def on_connected(self):
        """
        用作装饰器，可注册一个 CONNECTED 生命周期 hook 方法
        """
        return self.on(BotLife.CONNECTED, None)

    def on_before_close(self):
        """
        用作装饰器，可注册一个 BEFORE_CLOSE 生命周期 hook 方法
        """
        return self.on(BotLife.BEFORE_CLOSE, None)

    def on_before_stop(self):
        """
        用作装饰器，可注册一个 BEFORE_STOP 生命周期 hook 方法
        """
        return self.on(BotLife.BEFORE_STOP, None)

    def on_event_built(self):
        """
        用作装饰器，可注册一个 EVENT_BUILT 生命周期 hook 方法
        """
        return self.on(BotLife.EVENT_BUILT, None)

    def on_action_presend(self):
        """
        用作装饰器，可注册一个 ACTION_PRESEND 生命周期 hook 方法
        """
        return self.on(BotLife.ACTION_PRESEND, None)


BOT_PROXY = BotProxy()



import re

from ..types.abc import BotChecker
from ..types.exceptions import *
from ..types.typing import *

if TYPE_CHECKING:
    from ..models.event import MessageEvent, NoticeEvent, RequestEvent


class MsgLvlChecker(BotChecker):
    """
    消息分级权限校验器，不低于预设等级的 event 才会被接受。
    """

    def __init__(
        self,
        level: User,
        owner: int = None,
        super_users: List[int] = None,
        white_users: List[int] = None,
        black_users: List[int] = None,
    ) -> None:
        super().__init__()
        self.check_lvl = level

        self.owner = owner
        self.su_list = super_users if super_users is not None else []
        self.white_list = white_users if white_users is not None else []
        self.black_list = black_users if black_users is not None else []

    def _get_level(self, event: "MessageEvent") -> User:
        """
        获得事件对应的登记
        """
        qid = event.sender.id

        if qid in self.black_list:
            return User.BLACK
        elif qid == self.owner:
            return User.OWNER
        elif qid in self.su_list:
            return User.SU
        elif qid in self.white_list:
            return User.WHITE
        else:
            return User.USER

    def check(self, event: "MessageEvent") -> bool:
        """
        消息校验
        """
        e_level = self._get_level(event)
        return 0 < e_level.value and e_level.value >= self.check_lvl.value


class GroupMsgLvl(MsgLvlChecker):
    """
    群聊消息分级权限检查器，不低于预设等级的 event 才会被接受。
    不接受非白名单群聊的消息，也不接受任何私聊消息。
    特别注意：如果 white_groups 参数为 None，不启用白名单群聊校验
    """

    def __init__(
        self,
        level: User,
        owner: int = None,
        super_users: List[int] = None,
        white_users: List[int] = None,
        black_users: List[int] = None,
        white_groups: List[int] = None,
    ) -> None:
        super().__init__(level, owner, super_users, white_users, black_users)
        self.white_group_list = white_groups if white_groups is not None else []

    def check(self, event: "MessageEvent") -> bool:
        if not event.is_group():
            return False
        if len(self.white_group_list) == 0:
            return False
        if event.group_id not in self.white_group_list:
            return False
        return super().check(event)


class PrivateMsgLvl(MsgLvlChecker):
    """
    私聊消息分级权限检查器，不低于预设等级的 event 才会被接受。不接受任何群聊消息
    """

    def __init__(
        self,
        level: User,
        owner: int = None,
        super_users: List[int] = None,
        white_users: List[int] = None,
        black_users: List[int] = None,
    ) -> None:
        super().__init__(level, owner, super_users, white_users, black_users)

    def check(self, event: "MessageEvent") -> bool:
        if not event.is_private():
            return False
        return super().check(event)


class MsgCheckerGen:
    """
    消息校验器生成器。预先存储校验依据（各等级数据），
    指定校验 level 后返回一个符合 MsgLvlChecker 接口的实例对象
    """

    def __init__(
        self,
        owner: int = None,
        super_users: List[int] = None,
        white_users: List[int] = None,
        black_users: List[int] = None,
        white_groups: List[int] = None,
    ) -> None:
        self.owner = owner
        self.su_list = super_users if super_users is not None else []
        self.white_list = white_users if white_users is not None else []
        self.black_list = black_users if black_users is not None else []
        self.white_group_list = white_groups if white_groups is not None else []

    def gen_base(self, level: User = User.USER) -> MsgLvlChecker:
        """
        根据内部依据，生成一个原始消息等级校验器
        """
        return MsgLvlChecker(
            level, self.owner, self.su_list, self.white_list, self.black_list
        )

    def gen_group(self, level: User = User.USER) -> GroupMsgLvl:
        """
        根据内部依据，生成一个群组消息等级校验器
        """
        return GroupMsgLvl(
            level,
            self.owner,
            self.su_list,
            self.white_list,
            self.black_list,
            self.white_group_list,
        )

    def gen_private(self, level: User = User.USER) -> PrivateMsgLvl:
        """
        根据内部依据，生成一个私聊消息等级校验器
        """
        return PrivateMsgLvl(
            level, self.owner, self.su_list, self.white_list, self.black_list
        )


class AtChecker(BotChecker):
    """
    at 消息校验器。
    如果事件中包含指定 qid 的 at 消息，则校验结果为真
    """

    def __init__(self, qid: int = None) -> None:
        self.qid = str(qid) if qid is not None else None
        self._cq_at_regex = re.compile(r"\[CQ:at,qq=(\d+)?\]")

    def check(self, event: "MessageEvent") -> bool:
        """
        当 qid 为空时，只要有 at 消息就通过校验。
        如果不为空，则必须出现指定 qid 的 at 消息
        """
        id_list = self._cq_at_regex.findall(event.raw_content)
        if self.qid is None:
            return len(id_list) > 0
        for id in id_list:
            if id == self.qid:
                return True
        return False


class FriendReqChecker(BotChecker):
    """
    朋友请求事件校验器。
    如果事件是来自朋友的请求事件，则校验结果为真
    """

    def __init__(self) -> None:
        super().__init__()

    def check(self, event: "RequestEvent") -> bool:
        return event.is_friend_req()


class GroupReqChecker(BotChecker):
    """
    群请求事件校验器。
    如果事件是来自群的请求事件，则校验结果为真
    """

    def __init__(self) -> None:
        super().__init__()

    def check(self, event: "RequestEvent") -> bool:
        return event.is_group_req()


class NoticeTypeChecker(BotChecker):
    """
    通知事件类型校验器。
    校验是否为指定通知类型的通知事件
    """

    SUB_TYPES = [
        "group_upload",
        "group_admin",
        "group_decrease",
        "group_increase",
        "group_ban",
        "friend_add",
        "group_recall",
        "friend_recall",
        "group_card",
        "offline_file",
        "client_status",
        "essence",
        "notify",
        "honor",
        "poke",
        "lucky_king",
        "title",
        "ALL",
    ]

    def __init__(self, sub_type: str) -> None:
        super().__init__()
        if sub_type not in NoticeTypeChecker.SUB_TYPES:
            raise BotCheckerError(f"通知事件类型校验器的子类型 {sub_type} 不合法")
        self.sub_type = sub_type

    def check(self, event: "NoticeEvent") -> bool:
        if self.sub_type == "ALL":
            return True
        check_method = getattr(event, "is_" + self.sub_type)
        return check_method()



import ntpath
import os
import posixpath
import sys

import toml

from ..meta import EXIT_CLOSE
from ..types.typing import *
from .logger import get_logger


class ConfigGenerator:
    """
    配置生成器
    """

    def __init__(self, ConfigDirPath: str) -> None:
        self.config = {}
        self.default_config = DEFAULT_CONFIG
        self.default_config_text = DEFAULT_CONFIG_TEXT
        self.config_dir = ConfigDirPath
        self.config_path = os.path.join(self.config_dir, "melobot_config.toml")
        self.logger = get_logger()

    def get(self) -> dict:
        """
        获取配置文件
        """
        self.init_config()
        self.fill_defaults()
        return self.config

    def init_config(self) -> None:
        """
        初始化配置项，若没有配置文件，则创建。
        若存在配置文件，则读取配置
        """
        if not os.path.exists(self.config_dir):
            os.mkdir(self.config_dir)
        if not os.path.exists(self.config_path):
            with open(self.config_path, "w", encoding="utf-8") as fp:
                fp.write(self.default_config_text)
            self.logger.info("未检测到配置文件，已自动生成，请填写配置后重启 bot")
            exit(EXIT_CLOSE)
        else:
            with open(self.config_path, encoding="utf-8") as fp:
                self.config = toml.load(fp)

    def fill_defaults(self) -> None:
        """
        使用默认值填充不存在的配置项
        """
        for item_name in self.default_config.keys():
            if self.config.get(item_name) is None:
                self.config[item_name] = self.default_config[item_name]


class BotConfig:
    """
    配置类
    """

    def __init__(self, config_dir: str) -> None:
        self.connect_host: str = None
        self.connect_port: int = None
        self.max_conn_try: int = None
        self.conn_try_interval: int = None
        self.log_level: str = None
        self.cooldown_time: int = None
        self.log_dir_path: str = None

        self._build(config_dir)

    def _build(self, config_dir: str) -> None:
        self._raw = ConfigGenerator(config_dir).get()
        self.connect_host = self._raw["CONNECT_HOST"]
        self.connect_port = self._raw["CONNECT_PORT"]
        self.max_conn_try = self._raw["MAX_CONN_TRY"]
        self.conn_try_interval = self._raw["CONN_TRY_INTERVAL"]
        self.log_level = self._raw["LOG_LEVEL"]
        self.cooldown_time = self._raw["COOLDOWN_TIME"]
        self.log_dir_path = self._raw["LOG_DIR_PATH"]

        if sys.platform == "win32" and ntpath.isabs(self.log_dir_path):
            return
        if posixpath.isabs(self.log_dir_path):
            return
        self.log_dir_path = os.path.join(config_dir, self.log_dir_path)


DEFAULT_CONFIG_TEXT = """# 以下为自动生成的默认配置文件

# websocket 连接服务的 host
CONNECT_HOST = "localhost"
# websocket 连接服务的端口
CONNECT_PORT = 8080
# websocket 连接的最大重试次数（-1 为无限重试）
MAX_CONN_TRY = -1
# websocket 连接重试的间隔时间（-1 为无间隔）
CONN_TRY_INTERVAL = 2
# 全局日志等级（DEBUG, INFO, WARNING, ERROR, CRITICAL）
LOG_LEVEL = "INFO"
# 消息发送冷却时间（防止发送消息过快被风控）
COOLDOWN_TIME = 0.5
# 日志输出目录（会在该目录初始化 app.log 文件）
LOG_DIR_PATH = "../logs"
"""


DEFAULT_CONFIG = {
    "CONNECT_HOST": "localhost",
    "CONNECT_PORT": 8080,
    "MAX_CONN_TRY": -1,
    "CONN_TRY_INTERVAL": 2,
    "LOG_LEVEL": "INFO",
    "COOLDOWN_TIME": 0.5,
    "LOG_DIR_PATH": "../logs",
}



from setuptools import setup

from melobot.meta import MetaInfo

with open("requirements.txt", encoding="utf-8") as fp:
    required = fp.read().splitlines()
with open("README.md", encoding="utf-8") as fp:
    readme = fp.read()

META_INFO = MetaInfo()
setup(
    name="melobot",
    version=META_INFO.VER,
    author=META_INFO.AUTHOR,
    author_email=META_INFO.AUTHOR_EMAIL,
    description=META_INFO.PROJ_DESC,
    long_description=readme,
    long_description_content_type="text/markdown",
    python_requires=">=3.10",
    packages=[
        "melobot",
        "melobot.io",
        "melobot.context",
        "melobot.controller",
        "melobot.models",
        "melobot.plugin",
        "melobot.types",
        "melobot.utils",
    ],
    package_dir={
        "melobot": "melobot",
        "melobot.botio" : "melobot/io",
        "melobot.context" : "melobot/context",
        "melobot.controller" : "melobot/controller",
        "melobot.models" : "melobot/models",
        "melobot.plugin" : "melobot/plugin",
        "melobot.types" : "melobot/types",
        "melobot.utils" : "melobot/utils",
    },
    install_requires=required,
)



from .action import *
from .session import SESSION_LOCAL as session
from .session import (
    AttrSessionRule,
    BotSession,
    any_event,
    meta_event,
    msg_args,
    msg_event,
    msg_text,
    notice_event,
    pause,
    req_evnt,
)

session: BotSession



import asyncio as aio
import pathlib
import sys
import time
from asyncio import iscoroutine
from contextlib import asynccontextmanager
from functools import wraps

from ..types.exceptions import *
from ..types.typing import *


class Singleton:
    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "__instance__"):
            cls.__instance__ = super(Singleton, cls).__new__(cls)
        return cls.__instance__


class AsyncTwinEvent(aio.Event):
    """
    孪生 Event，会和绑定的一方时刻保持状态相反。
    """

    def __init__(self) -> None:
        super().__init__()
        self._twin: AsyncTwinEvent = None

    def bind(self, twin: "AsyncTwinEvent") -> None:
        self._twin = twin
        if self.is_set():
            super(AsyncTwinEvent, self._twin).clear()
        else:
            super(AsyncTwinEvent, self._twin).set()

    def set(self) -> None:
        super().set()
        if self._twin:
            super(AsyncTwinEvent, self._twin).clear()

    def clear(self) -> None:
        super().clear()
        if self._twin:
            super(AsyncTwinEvent, self._twin).set()


def get_twin_event() -> Tuple[aio.Event, aio.Event]:
    """
    获得两个时刻保持状态相反的 asyncio.Event。
    获得的第一个为 unset，另一个为 set
    """
    a, b = AsyncTwinEvent(), AsyncTwinEvent()
    a.bind(b)
    b.bind(a)
    return a, b


class RWController:
    """
    异步读写控制器。提供异步安全的读写上下文
    """

    def __init__(self, read_limit: int = None) -> None:
        @asynccontextmanager
        async def safe_read():
            nonlocal read_num, read_semaphore, write_semaphore, read_num_lock
            if read_semaphore:
                await read_semaphore.acquire()
            async with read_num_lock:
                if read_num == 0:
                    await write_semaphore.acquire()
                read_num += 1
            try:
                yield
            finally:
                async with read_num_lock:
                    read_num -= 1
                    if read_num == 0:
                        write_semaphore.release()
                    if read_semaphore:
                        read_semaphore.release()

        @asynccontextmanager
        async def safe_write():
            nonlocal write_semaphore
            await write_semaphore.acquire()
            try:
                yield
            finally:
                write_semaphore.release()

        write_semaphore = aio.Semaphore(1)
        if read_limit:
            read_semaphore = aio.Semaphore(read_limit)
        else:
            read_semaphore = None
        read_num = 0
        read_num_lock = aio.Lock()
        self.safe_read = safe_read
        self.safe_write = safe_write


class IdWorker:
    """
    雪花算法生成 ID
    """

    def __init__(self, datacenter_id, worker_id, sequence=0) -> int:
        self.MAX_WORKER_ID = -1 ^ (-1 << 3)
        self.MAX_DATACENTER_ID = -1 ^ (-1 << 5)
        self.WOKER_ID_SHIFT = 12
        self.DATACENTER_ID_SHIFT = 12 + 3
        self.TIMESTAMP_LEFT_SHIFT = 12 + 3 + 5
        self.SEQUENCE_MASK = -1 ^ (-1 << 12)
        self.STARTEPOCH = 1064980800000
        # sanity check
        if worker_id > self.MAX_WORKER_ID or worker_id < 0:
            raise ValueError("worker_id 值越界")
        if datacenter_id > self.MAX_DATACENTER_ID or datacenter_id < 0:
            raise ValueError("datacenter_id 值越界")
        self.worker_id = worker_id
        self.datacenter_id = datacenter_id
        self.sequence = sequence
        self.last_timestamp = -1

    def __gen_timestamp(self) -> int:
        """
        生成整数时间戳
        """
        return int(time.time() * 1000)

    def get_id(self) -> int:
        """
        获取新 ID
        """
        timestamp = self.__gen_timestamp()

        # 时钟回拨
        if timestamp < self.last_timestamp:
            raise ValueError(f"时钟回拨，{self.last_timestamp} 前拒绝 id 生成请求")
        if timestamp == self.last_timestamp:
            self.sequence = (self.sequence + 1) & self.SEQUENCE_MASK
            if self.sequence == 0:
                timestamp = self.__til_next_millis(self.last_timestamp)
        else:
            self.sequence = 0
        self.last_timestamp = timestamp
        new_id = (
            ((timestamp - self.STARTEPOCH) << self.TIMESTAMP_LEFT_SHIFT)
            | (self.datacenter_id << self.DATACENTER_ID_SHIFT)
            | (self.worker_id << self.WOKER_ID_SHIFT)
            | self.sequence
        )
        return new_id

    def __til_next_millis(self, last_timestamp) -> int:
        """
        等到下一毫秒
        """
        timestamp = self.__gen_timestamp()
        while timestamp <= last_timestamp:
            timestamp = self.__gen_timestamp()
        return timestamp


ID_WORKER = IdWorker(1, 1, 0)


def get_id() -> int:
    """
    获取一个全局唯一 id，由 melobot 内部 id 生成器提供
    """
    return ID_WORKER.get_id()


def this_dir(*relative_path: str) -> str:
    """
    包内脚本可通过该方法获取所在目录的绝对路径。提供 relative_path 参数，
    可自动拼接绝对路径。

    建议只在 py 脚本的顶级作用域调用，不要在其他作用域中调用
    """
    fr = sys._getframe(1)
    call_file = fr.f_locals["__file__"]
    return str(
        pathlib.Path(call_file).parent.joinpath(*relative_path).resolve(strict=True)
    )


def to_async(func: Callable):
    """
    异步包装器，将一个同步函数包装为异步函数。保留返回值。
    如果需要传参使用 partial 包裹
    """

    async def wrapper():
        return func()

    return wrapper


def to_coro(func: Callable):
    """
    协程包装器，将一个同步函数包装为协程。保留返回值。
    如果需要传参使用 partial 包裹
    """

    f = to_async(func)
    return f()


def lock(cb_maker: Callable[[None], Coroutine[Any, Any, Any]]) -> Callable:
    """
    锁装饰器，可以为被装饰的异步函数/方法加锁。
    在获取锁冲突时，调用 cb_maker 获得一个回调并执行。回调执行完毕后直接返回
    """
    alock = aio.Lock()
    if not callable(cb_maker):
        raise BotBaseUtilsError(
            f"lock 装饰器的 cb_maker 参数不可调用，cb_maker 值为：{cb_maker}"
        )

    def deco_func(func: Callable) -> Callable:
        @wraps(func)
        async def wrapped_func(*args, **kwargs):
            if alock.locked():
                cb = cb_maker()
                if not iscoroutine(cb):
                    raise BotBaseUtilsError(
                        f"lock 装饰器的 cb_maker 返回的不是协程，返回的回调为：{cb}"
                    )
                return await cb
            async with alock:
                return await func(*args, **kwargs)

        return wrapped_func

    return deco_func


def cooldown(
    busy_cb_maker: Callable[[None], Coroutine[Any, Any, Any]],
    cd_cb_maker: Callable[[float], Coroutine[Any, Any, Any]] = None,
    interval: float = 5,
) -> Callable:
    """
    冷却装饰器，可以为被装饰的异步函数/方法添加 cd 时间。
    cd_cb_maker 的类型：接受一个 float 参数（cd 剩余时间），
    返回一个协程的 Callable 对象。

    如果被装饰方法已有一个在运行，此时会直接调用 busy_cb_maker 生成一个回调并执行。
    回调执行完毕后直接返回。

    如果被装饰方法没有正在运行的，但在冷却完成前被调用，且此时 cd_cb_maker 不为 None，
    会使用 cd_cb_maker 生成一个回调并执行。如果此时 cd_cb_maker 为 None，
    被装饰方法会持续等待直至冷却结束再执行
    """
    alock = aio.Lock()
    pre_finish_t = time.time() - interval - 1
    if not callable(busy_cb_maker):
        raise BotBaseUtilsError(
            f"cooldown 装饰器的 busy_cb_maker 参数不可调用，busy_cb_maker 值为：{busy_cb_maker}"
        )
    if cd_cb_maker is not None and not callable(cd_cb_maker):
        raise BotBaseUtilsError(
            f"cooldown 装饰器的 cd_cb_maker 参数不可调用，cd_cb_maker 值为：{cd_cb_maker}"
        )

    def deco_func(func: Callable) -> Callable:
        @wraps(func)
        async def wrapped_func(*args, **kwargs):
            nonlocal pre_finish_t
            if alock.locked():
                busy_cb = busy_cb_maker()
                if not iscoroutine(busy_cb):
                    raise BotBaseUtilsError(
                        f"cooldown 装饰器的 busy_cb_maker 返回的不是协程，返回的回调为：{busy_cb}"
                    )
                return await busy_cb

            async with alock:
                duration = time.time() - pre_finish_t
                if duration > interval:
                    ret = await func(*args, **kwargs)
                    pre_finish_t = time.time()
                    return ret

                remain_t = interval - duration
                if cd_cb_maker is not None:
                    cd_cb = cd_cb_maker(remain_t)
                    if not iscoroutine(cd_cb):
                        raise BotBaseUtilsError(
                            f"cooldown 装饰器的 cd_cb_maker 返回的不是协程，返回的回调为：{cd_cb}"
                        )
                    return await cd_cb
                else:
                    await aio.sleep(remain_t)
                    ret = await func(*args, **kwargs)
                    pre_finish_t = time.time()
                    return ret

        return wrapped_func

    return deco_func


def semaphore(
    cb_maker: Callable[[None], Coroutine[Any, Any, Any]], value: int = -1
) -> Callable:
    """
    信号量装饰器，可以为被装饰的异步函数/方法添加信号量控制。
    在信号量无法立刻获取时，将调用 cb_maker 获得回调并执行。回调执行完毕后直接返回
    """
    a_semaphore = aio.Semaphore(value)
    if not callable(cb_maker):
        raise BotBaseUtilsError(
            f"semaphore 装饰器的 cb_maker 参数不可调用，cb_maker 值为：{cb_maker}"
        )

    def deco_func(func: Callable) -> Callable:
        @wraps(func)
        async def wrapped_func(*args, **kwargs):
            if a_semaphore.locked():
                cb = cb_maker()
                if not iscoroutine(cb):
                    raise BotBaseUtilsError(
                        f"semaphore 装饰器的 cb_maker 返回的不是协程，返回的回调为：{cb}"
                    )
                return await cb
            async with a_semaphore:
                return await func(*args, **kwargs)

        return wrapped_func

    return deco_func
